//  
//  (c) June 14, 2008 Vasily Volkov
//  

#pragma once

#define _USE_MATH_DEFINES
#include <math.h>

inline float2 operator*( float2 a, float2 b ) { return make_float2( a.x*b.x-a.y*b.y, a.x*b.y+a.y*b.x ); }
inline float2 operator*( float2 a, float  b ) { return make_float2( b*a.x, b*a.y ); }
inline float2 operator+( float2 a, float2 b ) { return make_float2( a.x + b.x, a.y + b.y ); }
inline float2 operator-( float2 a, float2 b ) { return make_float2( a.x - b.x, a.y - b.y ); }

//
// arrange blocks into 2D grid that fits into the GPU ( for powers of two only )
//
inline dim3 grid2D( int nblocks )
{
    int slices = 1;
    while( nblocks/slices > 65535 ) 
        slices *= 2;
    return dim3( nblocks/slices, slices );
}

//
//  exponentiation, bit reversal and basic FFT routines
//
#define exp_m18     make_float2(  1, -1 )//requires post-multiply by 1/sqrt(2)
#define exp_m14     make_float2(  0, -1 )
#define exp_m38     make_float2( -1, -1 )//requires post-multiply by 1/sqrt(2)

inline __device__ float2 exp_i( float phi )
{
    return make_float2( __cosf(phi), __sinf(phi) );
}

inline __device__ int rev( int bits )
{
    int reversed[] = {0,4,2,6,1,5,3,7};
    return reversed[bits];
}

inline __device__ void FFT2( float2 &a0, float2 &a1 )
{ 
    float2 c0 = a0;
    a0 = c0 + a1; 
    a1 = c0 - a1; 
}

//
//  loads
//
inline __device__ void load( float2 *a, float2 *x, int sx )
{
    for( int i = 0; i < 8; i++ )
        a[i] = x[i*sx];
}
inline __device__ void loadx( float2 *a, float *x, int sx )
{
    for( int i = 0; i < 8; i++ )
        a[i].x = x[i*sx];
}
inline __device__ void loady( float2 *a, float *x, int sx )
{
    for( int i = 0; i < 8; i++ )
        a[i].y = x[i*sx];
}

//
//  stores, input is in bit reversed order
//
inline __device__ void store( float2 *a, float2 *x, int sx )
{
    for( int i = 0; i < 8; i++ )
        x[i*sx] = a[rev(i)];
}
inline __device__ void storex( float2 *a, float *x, int sx )
{
    for( int i = 0; i < 8; i++ )
        x[i*sx] = a[rev(i)].x;
}
inline __device__ void storey( float2 *a, float *x, int sx )
{
    for( int i = 0; i < 8; i++ )
        x[i*sx] = a[rev(i)].y;
}

//
// FFT(a[0:7]), output is in bit reversed order
//
inline __device__ void FFT8( float2 *a )
{
    //ffts (16 adds)
    FFT2( a[0], a[4] );
    FFT2( a[1], a[5] );
    FFT2( a[2], a[6] );
    FFT2( a[3], a[7] );
    
    //twiddles (4 adds, 4 muls)
    a[5] = ( a[5] * exp_m18 ) * M_SQRT1_2;
    a[6] =   a[6] * exp_m14;
    a[7] = ( a[7] * exp_m38 ) * M_SQRT1_2;
    
    //ffts (16 adds)
    FFT2( a[0], a[2] );
    FFT2( a[1], a[3] );
    FFT2( a[4], a[6] );
    FFT2( a[5], a[7] );
    
    //twiddles (0 flops)
    a[3] = a[3] * exp_m14;
    a[7] = a[7] * exp_m14;
    
    //ffts (16 adds)
    FFT2( a[0], a[1] );
    FFT2( a[2], a[3] );
    FFT2( a[4], a[5] );
    FFT2( a[6], a[7] );
}

//
//  multiply by twiddle factors in bit-reversed layout
//
inline __device__ void twiddle( float2 *a, int i, int n )
{
#pragma unroll
    for( int j = 1; j < 8; j++ )
        a[j] = a[j] * exp_i((-2*M_PI*rev(j)/n)*i);
}

//
//  transpose via shared memory, input is in bit-reversed layout
//
inline __device__ void tranpose( float2 *a, float *s, int ds, float *l, int dl, int sync = 0xf )
{
    storex( a, s, ds );	if( sync&8 ) __syncthreads();
    loadx ( a, l, dl );	if( sync&4 ) __syncthreads();
    storey( a, s, ds );	if( sync&2 ) __syncthreads();
    loady ( a, l, dl );	if( sync&1 ) __syncthreads();
}
